#include <fcntl.h>
#include <math.h>

#include "psi_obj.h"
#include "psi_term.h"
#include "exc.h"
#include "obj_file.h"
#include "obj_page.h"
#include "obj_stream.h"
#include "o_buffer.h"
#include "db.h"
#include "defs.h"

extern "C" {
#include "Source/memory.h"
#include "Source/trees.h"
#include "Source/print.h"
#include "Source/login.h"
}

#define BYTE_COPY(a, b, l) bcopy((const char*)(b), (char*)(a), l)

#define deref_s_ptr(P, Prev) \
      ptr_psi_term Prev; \
      Prev=NULL; \
      while(P->coref) { Prev=P; P=P->coref; }

#define LARGE_ENOUGH(C) \
      if (!C) {\
      dberror << "sorry large object storage not yet implemented"; \
      throw_error(); \
      }

#define IS_LARGE_OBJECT LARGE_ENOUGH( (object_size <= OBJ_SIZE_MAX) )

void string2number(ptr_psi_term&, char*);

int save_level = -1;

void po_temp() {

}

//////////////////////
// CLASS PsiObject //
////////////////////

PsiObject:: PsiObject(ptr_psi_term term, ObjFile* file,
		      int near_hint):obj_page(NULL)
				// term is stored in file.
{  
  UNEXPECTED( (term->dbpointer),
	      "unexpected error (dbpointer) in PsiObject:: PsiObject(ptr_psi_term term, ObjFile* file, int near_hint)");

  if (file->file_writable()) {
    page_num pnum = 
      (near_hint==START)? file->first_obj_page():
      file->last_obj_page();
    
    PageKey p_key(file, pnum);
    obj_page = new ObjectPage(p_key); 
    
    if (save_top_only) save_level=1;
    else save_level=-1;
    psi_term2object(term);
  }
}

PsiObject:: PsiObject(ptr_psi_term term, slot_num snum, 
		      ObjectPage& page):obj_page(NULL)
				// term is updated tp the page
{

  if (page.file()->file_writable()) {
    obj_page = new ObjectPage(page.key());
    
    if (save_top_only) save_level=1;
    else save_level=-1;
    psi_term2object(term, snum, page); 
         // obj_page value might change if sufficient space is not left 
        // in obj_page. term was previously stored in page. it might get
       // forwarded to a different obj_page
        // In the function 
       // psi_term2object(term, snum, page) page should be fixed.
  }

}

PsiObject:: PsiObject(ptr_psi_term term):obj_page(NULL)
				// term is updated
{
  UNEXPECTED( (!term->dbpointer),
	      "unexpected error (no dbpointer) in PsiObject:: PsiObject(ptr_psi_term term)" );
  ObjectId& o_id = PSIID(term->dbpointer);      
  if (o_id.file()->file_writable()) {
    ObjectPage page(o_id);
    obj_page = new ObjectPage(o_id);      
    // obj_page could change to a different value.
    if (save_top_only) save_level=1;
    else save_level=-1;
    psi_term2object(term, o_id.snum(), page);
    // obj_page value might change if sufficient space is not left 
    // in obj_page. term was previously stored in page. it might get
    // forwarded to a different obj_page
    // In the function 
    // psi_term2object(term, snum, page) page should be fixed.
  }
}

PsiObject:: ~PsiObject()  { delete obj_page; }

static char real_buffer[REAL_STRLEN];

void PsiObject:: sort_info(ptr_psi_term term, char*& sort_value, 
			     u_1byte& sort_type, 
			     unsigned long& sort_size) {

  deref_ptr(term);
  if (term->value) {
    if (sub_type(term->type,real)) {
//       sort_value = (char *)term->value;
//       sort_size = sizeof(REAL);
//       sort_type = REAL_TYPE;

      sprintf(real_buffer,"%g",*(REAL *)term->value);
      sort_value = real_buffer;
      sort_size = strlen(sort_value)+1;
      sort_type = REAL_TYPE;
    }
    else if (sub_type(term->type,quoted_string)) {
      sort_value = (char*)term->value;
      sort_size = strlen(sort_value)+1;
      sort_type = STRING_TYPE;
    }
    else {
      sort_value = term->type->keyword->combined_name;
      sort_size = strlen(sort_value)+1;
      sort_type = SORT_TYPE;
    }
  }
  else {			// if (!term->value)
    sort_value = term->type->keyword->combined_name;
    sort_size = strlen(sort_value)+1;
    sort_type = SORT_TYPE;
  };
}

void PsiObject:: sort2char_1(char*& dp, char* sort_value, 
			     u_1byte sort_type, unsigned long len) {

  object_size += (OI_SIZE+TI_SIZE+SI_SIZE+len);
  IS_LARGE_OBJECT;
  LARGE_ENOUGH( (len <= SI_MAX) );

  dp += OI_SIZE;
  *(u_1byte*)dp = sort_type;
  dp += TI_SIZE;
  u_2byte sort_size = (u_2byte)len;
  cp_u_2byte_to_pt(dp, sort_size);
  dp += SI_SIZE;
  BYTE_COPY(dp, sort_value, len);
  dp += len;
}

void PsiObject:: sort2char_2(char*& dp, char* sort_value, 
			      u_1byte sort_type, unsigned long len) {

  object_size += (OI_SIZE+TI_SIZE+len);
  IS_LARGE_OBJECT;

  dp += OI_SIZE;
  *(u_1byte*)dp = sort_type;
  dp += TI_SIZE;
  BYTE_COPY(dp, sort_value, len);
  dp += len;
}

int PsiObject:: psi_tree2objectA(ptr_node n, char* dp,
				 unsigned long& num_of_fields) 
{
  // dp points to "# of fields" in the object layout on disk

  num_of_fields=0;
  unsigned long pts_fields_length = 0;
  int incomplete_object=
    psi_tree2objectA_2(n, num_of_fields, pts_fields_length);

  unsigned long field_size = num_of_fields*FI_SIZE;
  LARGE_ENOUGH( (num_of_fields <= FNUM_MAX) );
  object_size += (FNUM_SIZE + field_size + pts_fields_length);
  IS_LARGE_OBJECT;


  u_2byte temp = 0;
  cp_u_2byte_to_pt(dp,  temp);
  //  cerr << "incomplete_object=" << incomplete_object << "\n";
  return incomplete_object;

}

int PsiObject:: psi_tree2objectA_2(ptr_node tree, 
				   unsigned long& num, unsigned long& len) 
{
  int incomplete_object=0;
  if(tree) {
    if(tree->left)
      incomplete_object=psi_tree2objectA_2(tree->left, num, len);
    if (tree->data) {
      num++;
      len += strlen(tree->key)+1;
      ptr_psi_term term = (ptr_psi_term)tree->data;
      deref_ptr(term);
      if (term != root_sort) {
	len += OID_SIZE;
	if (!incomplete_object) incomplete_object=(term->dbpointer?0:1);
      }
    }
    if(tree->right)
      if (psi_tree2objectA_2(tree->right, num, len))
	incomplete_object=1;
  }
  return incomplete_object;
}

void PsiObject:: psi_tree2objectB(ptr_node tree, char* fp, 
				  unsigned long num_of_fields,
				  ptr_psi_term root) 
{
  //  unsigned long num_of_fields = PsiTerm:: num_of_features(tree);
  u_2byte temp = num_of_fields;
  cp_u_2byte_to_pt(fp,  temp);
  fp += FNUM_SIZE;
  unsigned long field_size = num_of_fields*FI_SIZE;
  char* dp = fp+field_size;
  deref_ptr(root);
  psi_tree2objectB_2(tree, dp, fp, root); 
}

void PsiObject:: psi_tree2objectB_2(ptr_node tree, char*& dp, char*& fp,
				   ptr_psi_term root){

  if(tree) {
    if(tree->left)
      psi_tree2objectB_2(tree->left, dp, fp, root);
    
    if (tree->data) {
      unsigned long index = dp - fp;
      LARGE_ENOUGH( (index < FI_MAX) );
      u_2byte temp = index;
      cp_u_2byte_to_pt(fp, temp);

      ptr_psi_term term = (ptr_psi_term)tree->data;
      deref_ptr(term);

      unsigned long len = strlen(tree->key) + 1;
      if (term==root) {
	*(u_1byte*)fp |= FI_LOOP_MASK;
	fp += FI_SIZE;
	
	BYTE_COPY(dp, tree->key, len);
	dp += len;
	// cerr << "key: " << tree->key << "\n";
      }
      else {
	fp += FI_SIZE;
	
	if (!term->dbpointer) { // this will avoid looping also in case 
	             // save level is infinity (that is  a negative value)
	  save_level--;
	  psi_term2object(term);
	  save_level++;
	  deref_ptr(term);
	}
	ObjectId& object_id = PSIID(term->dbpointer);
	cp_p_4byte_to_pt(dp, object_id.pnum());
	dp += 4;
	cp_s_2byte_to_pt(dp, object_id.snum());
	dp += 2;
	BYTE_COPY(dp, tree->key, len);
	dp += (len);

// 	page_num pnum = object_id.pnum();
// 	slot_num snum = object_id.snum();

// 	cerr << "key:" << tree->key;
// 	cerr << " pnum:" << pnum << " snum:" << snum << "\n";

// 	if ( !pnum || !snum ) {
// 	  cerr << "unexpected null values";
// 	}


      }

    }

    if(tree->right)
      psi_tree2objectB_2(tree->right, dp, fp, root);
  }
}

void PsiObject:: psi_term2object(ptr_psi_term term) {

  deref_ptr(term);

  UNEXPECTED( (term->type==psi_id),
	      "unexpected error in PsiObject::psi_term2object" );
  UNEXPECTED( (term->dbpointer),
	     "unexpected error (dbpointer exist) in PsiObject::psi_term2object(ptr_psi_term term, slot_num snum, ObjectPage& page)" );

  object_size = 0;
  char data[OBJ_SIZE_MAX];
  char*  dp = data;

  ptr_node node = (ptr_node) term->attr_list;

  char* sort_value;
  u_1byte sort_type;
  unsigned long len;
  o_2byte size = 0;

  sort_info(term, sort_value, sort_type, len);
  if (node) {
    sort_type |= PSI_MASK;
    sort2char_1(dp, sort_value, sort_type, len);
    root_sort=term;
    deref_ptr(root_sort);
    unsigned long num_of_fields;
    int incomplete_object = 
      psi_tree2objectA(node, dp, num_of_fields);
    
    char* field_index = dp;
    dp = data+OI_SIZE;
    *(u_1byte*)dp |= check_object_size();

    size = object_size;
    cp_o_2byte_to_pt(data, size);
    create_object(term, size, data);
    deref_ptr(term);
    ObjectId& object_id = PSIID(term->dbpointer);
    if (save_level == 0 && incomplete_object) {
      object_id.incomplete_object();
    }
    else {
      psi_tree2objectB(node, field_index, num_of_fields, root_sort);
      update_object(object_id, size, data);
    }
  }
  else {
    sort2char_2(dp, sort_value, sort_type, len);

    dp = data+OI_SIZE;
    *(u_1byte*)dp |= check_object_size();
    size = object_size;
    cp_o_2byte_to_pt(data, size);
    create_object(term, size, data);
  }
}

void PsiObject:: psi_term2object(ptr_psi_term term, slot_num snum,
				  ObjectPage& page) 
{
  deref_ptr(term);
  UNEXPECTED( (term->type==psi_id),
	     "unexpected error in PsiObject::psi_term2object(ptr_psi_term term, slot_num snum, ObjectPage& page)" );

  UNEXPECTED( (!term->dbpointer),
	     "unexpected error (no dbpointer) in PsiObject::psi_term2object(ptr_psi_term term, slot_num snum, ObjectPage& page)" );

  ObjectId& object_id = PSIID(term->dbpointer);
  object_id.updated_psi_term();
  object_size = 0;
  char data[OBJ_SIZE_MAX];
  char*  dp = data;

  ptr_node node = (ptr_node) term->attr_list;

  char* sort_value;
  u_1byte sort_type;
  unsigned long len;
  o_2byte size = 0;

  sort_info(term, sort_value, sort_type, len);

  if (node) {
    sort_type |= PSI_MASK;
    sort2char_1(dp, sort_value, sort_type, len);
    root_sort=term;
    deref_ptr(root_sort);
    unsigned long num_of_fields;
    int incomplete_object = psi_tree2objectA(node, dp, num_of_fields);
    
    char* field_index = dp;
    dp = data+OI_SIZE;
    *(u_1byte*)dp |= check_object_size();

    size = object_size;
    cp_o_2byte_to_pt(data, size);
    
    slot_num snum_f;
    if (snum_f=page.realloc_space(snum,size)) {
      if (save_level == 0 && incomplete_object) {
	object_id.incomplete_object();
      }
      else {
	psi_tree2objectB(node, field_index, num_of_fields, root_sort);
      }
      page.update_object(snum_f, size, data);
    }
    else {
      OId o_id;   
      //      create_forwarded_object(o_id, snum, size, data);
      //      page.modify_object(snum, o_id.pnum_d, o_id.snum_d, F_TYPE);

     if (save_level == 0 && incomplete_object) {
	object_id.incomplete_object();
      }
      else {

	create_forwarded_object(o_id, snum, size, data);

	page.modify_object(snum, o_id.pnum_d, o_id.snum_d, F_TYPE);

	psi_tree2objectB(node, field_index, num_of_fields, root_sort);
	update_object(o_id, size, data);
      }
    }

  }
  else {
    sort2char_2(dp, sort_value, sort_type, len);

    dp = data+OI_SIZE;
    *(u_1byte*)dp |= check_object_size();
    size = object_size;
    cp_o_2byte_to_pt(data, size);
    
    if (!page.modify_object(snum,size,data)) {
      create_forwarded_object(page, snum, size, data);
    }
  }
}

void PsiObject:: sufficient_space(o_2byte size) {
  if ( !obj_page->has_space(size) ) {
    ObjectPage* old_page = obj_page;
    page_num pnum = obj_page->last_obj_page();
    if ( pnum == obj_page->pnum() )
      obj_page = new ObjectPage(NewPage(), *old_page);
    else {
      PageKey p_key(obj_page->file(), pnum);
      obj_page = new ObjectPage(p_key);
      if (!obj_page->has_space(size)) {
	//	cerr << "size2: " << size << " " << *obj_page << "\n";
	delete obj_page;
	obj_page = new ObjectPage(NewPage(), *old_page);
      }
    }
    delete old_page;
  }
}

void PsiObject:: create_object(ptr_psi_term term, o_2byte size, 
			       const char* data) {

  sufficient_space(size);
  slot_num snum = obj_page->create_object(size, data);
  ObjectKey o_key(obj_page->file(), obj_page->pnum(), snum); 
  object_stream->object_buffer()->insert_psi_term(o_key, term);
}

void PsiObject:: update_object(const ObjectKey& o_key, o_2byte size,
				const char* data) {
  if (obj_page->same_page(o_key))
      obj_page->update_object(o_key.snum(), size, data);
  else {
    delete obj_page;
    obj_page = new ObjectPage(o_key);
    obj_page->update_object(o_key.snum(), size, data);
  }
}

void PsiObject:: create_forwarded_object(OId& o_id, slot_num snum, 
					 o_2byte size, const char* data) 
{
  sufficient_space(size);
  o_id.file_d = obj_page->file();
  o_id.pnum_d = obj_page->pnum();
  o_id.snum_d = obj_page->create_object(size, data);
}

void PsiObject:: create_forwarded_object(ObjectPage& page, slot_num snum,   
					 o_2byte size, const char* data) 
{
  sufficient_space(size);
  slot_num new_snum = obj_page->create_object(size,data);
  page.modify_object(snum, obj_page->pnum(), new_snum, F_TYPE);
}

u_1byte PsiObject:: check_object_size() {
  u_1byte pos = POINTER_OBJECT_SIZE;
  if (object_size < pos) {
    u_1byte diff = pos-object_size;
    UNEXPECTED( (diff > PADDED_SIZE_MAX),
	       "increase the PADDED_SIZE_MAX in PsiObject:: check_object_size()" );
    object_size=pos;
    IS_LARGE_OBJECT;
    return diff;
  }
  return 0;
}

//////////////////////
// CLASS PsiIdTerm //
////////////////////
  // the hashed term on o_key is returned otherwise a new heap term is 
  // is created hashed on o_key.
#ifdef   CHECK_UNEXPECTED
static int loop_d=0;
PsiIdTerm:: PsiIdTerm(const ObjectKey& o_key, psi_loop&)
                    
{
  loop_d++; 

  if ( term_d = object_stream->object_buffer()->get_psi_term(o_key) ) {
  }
  else {
    term_d = object_stream->object_buffer()->insert_psi_term(o_key);
    deref_ptr(term_d);
    UNEXPECTED( (!term_d && !term_d->dbpointer), 
		"unexpected error(3) in PsiIdTerm:: PsiIdTerm(const ObjectKey& o_key)-- term not properly hashed");
  }
}
#endif

PsiIdTerm:: PsiIdTerm(const ObjectKey& o_key)
{

#ifdef   CHECK_UNEXPECTED
  loop_d=0;
#endif

  if ( term_d = object_stream->object_buffer()->get_psi_term(o_key) ) {
  }
  else {
    term_d = object_stream->object_buffer()->insert_psi_term(o_key);
    deref_ptr(term_d);
    UNEXPECTED( (!term_d && !term_d->dbpointer), 
		"unexpected error(3) in PsiIdTerm:: PsiIdTerm(const ObjectKey& o_key)-- term not properly hashed");
  }
}

PsiIdTerm:: PsiIdTerm(ptr_psi_term term):term_d(term) {

#ifdef   CHECK_UNEXPECTED
  loop_d=0;
#endif

  deref_ptr(term_d);

  UNEXPECTED( (!term_d->dbpointer),
	      "unexpected error in PsiIdTerm:: PsiIdTerm(ptr_psi_term term)");

  ObjectId& object_id = PSIID(term_d->dbpointer);
  if (!object_id.is_hashed()) {
    ptr_psi_term term2 = object_stream->object_buffer()->get_psi_term(object_id);
    if (term2) {
      deref_ptr(term2);
      term_d->coref=term2;
    }
    else
      object_stream->object_buffer()->insert_psi_term(term_d);
  }
}

void PsiIdTerm:: swizzle() {

  deref_ptr(term_d);
  if ( term_d->type==psi_id && term_d->dbpointer) {
    ObjectId& object_id = PSIID(term_d->dbpointer);

    Object2Psi obj(object_id);

    ptr_psi_term new_term = obj;
    if (new_term) {
      deref_ptr(new_term);
      new_term->dbpointer=term_d->dbpointer;
      term_d->coref=new_term;
    }
    else {
      // it is a pointer to another object
      // cerr << "correfferenced term\n";
      ObjectKey o_key(object_id.file(), obj.coref_pnum, obj.coref_snum);
      UNEXPECTED( (loop_d > 1000),
	"unexpected looping in PsiIdTerm:: swizzle()");
      //      if (loop_d > 20) {
      //	cerr << object_id << " is coreffed to "
      //	     << "pnum: " << obj.coref_pnum 
      //	     << " snum: " << obj.coref_snum << "\n";
      //      }
#ifdef   CHECK_UNEXPECTED
      psi_loop looping;
      PsiIdTerm id(o_key, looping); // psiidterm call
#else
      PsiIdTerm id(o_key);
#endif
      id.swizzle();
      new_term=id;
      deref_ptr(new_term);
      term_d->coref=new_term;
      if (new_term->dbpointer) 
	object_id.dbcoreffed_psi_term(new_term->dbpointer);
    }

    deref_ptr(term_d);

    UNEXPECTED( (!term_d->dbpointer),
		"unexpected null value in PsiIdTerm:: swizzle()"  );
    UNEXPECTED( (object_id.dbpointer() != term_d->dbpointer),
		"unexpected error in PsiIdTerm:: swizzle()"  );
  }
}
  
// void PsiIdTerm:: feature_added() {
//   deref_ptr(term_d);
//   if (term_d->dbpointer) {
//     ObjectId& object_id = PSIID(term_d->dbpointer);
//     object_id.feature_added2psi_term();
//   }
// }


//////////////////////
// class Object2Psi //
////////////////////
Object2Psi:: operator ptr_psi_term() {
  if (term_d)  deref_ptr(term_d);
  return term_d;
}

float Object2Psi:: swizzle_time=0;
long Object2Psi:: times_swizzle_called=0;

ptr_psi_term Object2Psi:: object2psi_term(ObjectKey& o_key){
  ptr_psi_term term=NULL;

  ObjectPage page(o_key);

  if (o_key.snum() > page.num_of_objects()) {
    dberror << "invalid object reference: (" << o_key.file()->name() 
      << ", " << o_key.pnum() << ", " << o_key.snum() << ")";
    throw_error();
  }

  struct tms swizzle_start_time, swizzle_end_time;
  times(&swizzle_start_time);
  const char* dp = page[o_key.snum()];  
  page.ptrs_lock(1);
  u_1byte sort_type;
  u_2byte sort_size;
  const char* sort_value;

  int has_fields = sort_info(dp, sort_value, sort_type, sort_size);
  char2sort(term, sort_value, sort_type, sort_size);
  if (term) {
    deref_ptr(term);
    root_sort = term;
    UNEXPECTED ( (term->type==psi_id),
		 "unexpected error in Object2Psi:: object2psi_term(ObjectKey& o_key)");
    //  if (o_key.pnum()==60 && o_key.snum()==2) po_temp();
    if (has_fields) {
      object_tree(term->attr_list, dp);
    }
  } // else a pointer to another object

  page.ptrs_lock(0);
  times(&swizzle_end_time);
  float time_taken = (swizzle_end_time.tms_utime - 
		      swizzle_start_time.tms_utime);
  swizzle_time+=time_taken;
  times_swizzle_called++;

  return term;
}

void Object2Psi:: object_tree(ptr_node& node, const char* dp) {
  u_2byte num_of_fields;
  pt2u_2byte(num_of_fields, dp);
  //  cerr << "nof:" << num_of_fields << "\n";
  dp += FNUM_SIZE;
  object_tree_2(node, dp, num_of_fields);
}

void Object2Psi::object_tree_2(ptr_node& node, 
			    const char* dp, u_2byte num) {
  if (num > 0) {
    u_2byte mid = (num-1) >> 1;
    //    cerr << "mid:" << mid << "\n";
    const char* data_ptr = dp + mid*FI_SIZE;

    char i[FI_SIZE];
    BYTE_COPY(i, data_ptr, FI_SIZE);
    int is_loop = i[0] & FI_LOOP_MASK;
    
    i[0] &= FI_SIZE_MASK;
    u_2byte index;
    pt2u_2byte(index, i);
    data_ptr += (index);
    
    if (is_loop) {
      //      cerr << "loop_key:" << data_ptr;
      heap_insert_copystr((char*)data_ptr, &node, root_sort);
    }
    else {
      ptr_psi_term term;
      char2psi_id(term, data_ptr);
      heap_insert_copystr(data_ptr, &node, term);
    }      
    
    object_tree_2(node, dp, mid);
    object_tree_2(node, dp+(mid+1)*FI_SIZE, num-mid-1);
  }
}

void Object2Psi:: char2sort(ptr_psi_term& term, const char* dp, 
			     u_1byte sort_type, u_2byte len) { 
  switch (sort_type) {

  case SORT_TYPE:
    term = heap_psi_term(4);
    term->type=PsiTerm:: update_symbol(NULL, dp);
    break;
  case STRING_TYPE:
    {
      term = heap_psi_term(4);
      term->type=quoted_string;
      char* sort_value=(char*)heap_alloc(len);
      BYTE_COPY(sort_value, dp, len);
      term->value=(GENERIC)sort_value;
    }
    break;
  case REAL_TYPE:
    {
//       term = heap_psi_term(4);
//       char* sort_value=(char*)heap_alloc(len);
//       BYTE_COPY(sort_value, dp, len);
//       term->value=(GENERIC)sort_value;
//       REAL v = *(REAL *)term->value;
//       if (v==floor(v))
// 	term->type=integer;
//       else
// 	term->type=real;

      string2number(term, dp);
    }
    break;
  case F_TYPE:
    {				//  forwarded object in same file
      page_num pnum;
      slot_num snum;
      pt2p_4byte(pnum, dp);
      dp += 4;
      pt2s_2byte(snum, dp);
      
      ObjectKey o_key(file_d, pnum, snum);

      Object2Psi obj(o_key);

      term = obj;
      UNEXPECTED( (!term),
		  "unexpected error(1) in Object2Psi:: char2sort()");
      // a pointer object can never be forwarded, otherwise 
      // coref_pnum and coref_snum have to be made static
    }
    break;
  case OID_TYPE:
    {		       //  object points to another object in same file
      //      page_num pnum;
      //      slot_num snum;
      //      pt2p_4byte(pnum, dp);
      //      dp += 4;
      //      pt2s_2byte(snum, dp);
      //PsiIdTerm id(ObjectKey(file_d, pnum, snum), 1);
      //      term=id;
      
      term=NULL;
      pt2p_4byte(coref_pnum, dp);
      dp += 4;
      pt2s_2byte(coref_snum, dp);
    }
    break;
  default:
    UNEXPECTED( (1),
		"unexpected error in Object2Psi:: char2sort()");
    term = heap_psi_term(4);
    break;
  }

}

int Object2Psi:: sort_info(const char*& obj, const char*& sort_value,
			  u_1byte& sort_type, u_2byte& sort_size) {
  int has_fields = 0;
  sort_type = *(u_1byte*)(obj+OI_SIZE); 
  if ( sort_type & PSI_MASK ) { 
    obj += (OI_SIZE+TI_SIZE);
    has_fields=1;
    pt2u_2byte( sort_size, obj ); 
    obj += SI_SIZE;
  } 
  else { 
    pt2u_2byte( sort_size, obj ); 
    u_1byte padded_size = sort_type & PADDED_MASK;
    sort_size -= (OI_SIZE+TI_SIZE+padded_size); 
    obj += (OI_SIZE+TI_SIZE); 
  }; 
  sort_type &= SORT_TYPE_MASK;
  sort_value = obj; 
  obj += sort_size;
  return has_fields;
}

void Object2Psi:: char2psi_id(ptr_psi_term& term, 
			     const char*& data_ptr) {
  page_num pnum;
  slot_num snum;

  pt2p_4byte(pnum, data_ptr);
  data_ptr += 4;
  pt2s_2byte(snum, data_ptr);
  data_ptr += 2;

  //  cerr << "key:" << data_ptr;
  //  cerr << " pnum:" << pnum << " snum:" << snum << "\n";

  if ( !pnum || !snum ) {
    dberror << "unexpected null-values in file " << file_d->name();
    throw_error();
  }

  ObjectKey o_key(file_d,pnum,snum);
  PsiIdTerm id(o_key); //psiidterm call
  term = id;
}

void Object2Psi:: output_data(ostream& os){
//  ptr_psi_term term = object2psi_term();
//  cout << term;
}

ostream& operator<<(ostream& os, Object2Psi& obj) {
  obj.output_data(os);
  return os;
}

////////////////////////////////////////////////////////////////
/******** STRING2NUMBER(c)
  Accepted syntax: digit+ [ . digit+ ] [ {e|E} {+|-|empty} digit* ]
*/
void string2number(ptr_psi_term& term, char* real_string)
{
  term = heap_psi_term(4);

  REAL f,p;
  long sgn,pwr,posflag;

  char* real_ptr = real_string;
  if (sgn=(*real_ptr=='-')) real_ptr++; 

  f=0.0;
  do { f=f*10.0+(*real_ptr-'0'); real_ptr++; } while (DIGIT(*real_ptr));

  if (*real_ptr=='.') {
    real_ptr++;
    if DIGIT(*real_ptr) {
      /* term->type=real;     RM: Mar  8 1993  */
      p=10.0;
      while (DIGIT(*real_ptr)) { 
	f=f+(*real_ptr-'0')/p; p=p*10.0; real_ptr++; 
      }
      real_ptr--;
    }
    else {
      real_ptr--;
      real_ptr--;
    }
  }
  else
    real_ptr--;

  real_ptr++;
  if (*real_ptr=='e' || *real_ptr=='E') {
    real_ptr++;
    if (*real_ptr=='+' || *real_ptr=='-' || DIGIT(*real_ptr)) {
      term->type=real;
      posflag = (*real_ptr=='+' || DIGIT(*real_ptr));
      if (!DIGIT(*real_ptr)) real_ptr++;;
      pwr=0;
      while (DIGIT(*real_ptr)) { pwr=pwr*10+(*real_ptr-'0'); real_ptr++; }
      real_ptr--;
      p=1.0;
      while (pwr>=100) { pwr-=100; if (posflag) p*=1e100; else p/=1e100; }
      while (pwr>=10 ) { pwr-=10;  if (posflag) p*=1e10;  else p/=1e10;  }
      while (pwr>0   ) { pwr-=1;   if (posflag) p*=1e1;   else p/=1e1;   }
      f*=p;
    }
    else {
      real_ptr--;
      real_ptr--;
    }
  }
  else
    real_ptr--;

  if (sgn) f = -f; 
  term->value=heap_alloc(sizeof(REAL)); /* 12.5 */
  *(REAL *)term->value=f;

  /*  RM: Mar  8 1993  */
  if(f==floor(f))
    term->type=integer;
  else
    term->type=real;
}

static void fetch_db_psi_node(ptr_node tree, int level){

  if(tree) {
    if(tree->left)
      fetch_db_psi_node(tree->left, level);
    
    if (tree->data) {
      ptr_psi_term term = (ptr_psi_term)tree->data;
      deref_ptr(term);
      if (term->type==psi_id && term->dbpointer)
	fetch_db_psi_term(term, level);
    }

    if(tree->right)
      fetch_db_psi_node(tree->right, level);
  }

}

void fetch_db_psi_term(ptr_psi_term& term, int level) {
  {
    PsiIdTerm id(term); // psiidterm call
    id.swizzle();
    deref_ptr(term);
    if (level > 0) {
//      cerr << "level: " << level << "\n";
      fetch_db_psi_node(term->attr_list, level-1);
    }
    
  }
}

